# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e
import copy

gconfig = {}
gconfig = TestBaseClass.get_connection_config()


def test_setting_key():
    key_val = aerospike.POLICY_KEY_SEND
    read_policy = {"key": key_val}
    policies = {"read": read_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)

    aerospike.client(config)


# TODO: duplicate test name
@pytest.mark.xfail(reason="Had a duplicate test name, but fails when renamed.")
def test_setting_consistency_duplicate():
    cons_val = aerospike.POLICY_CONSISTENCY_ONE
    read_policy = {"consistency_level": cons_val}
    policies = {"read": read_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    aerospike.client(config)


def test_setting_consistency():
    replica_val = aerospike.POLICY_REPLICA_MASTER
    read_policy = {"replica": replica_val}
    policies = {"read": read_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    aerospike.client(config)


def test_setting_conmmit_level():
    commit_val = aerospike.POLICY_COMMIT_LEVEL_ALL
    write_policy = {"commit_level": commit_val}
    policies = {"write": write_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    aerospike.client(config)


def test_setting_exists():
    exists_val = aerospike.POLICY_EXISTS_CREATE
    write_policy = {"commit_level": exists_val}
    policies = {"write": write_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    aerospike.client(config)


def test_setting_gen():
    gen_val = aerospike.POLICY_GEN_IGNORE
    write_policy = {"commit_level": gen_val}
    policies = {"write": write_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    aerospike.client(config)


def test_setting_wrong_type():
    write_policy = {"commit_level": [1, 2, 3]}
    policies = {"write": write_policy}
    config = copy.deepcopy(gconfig)
    config["policies"].update(policies)
    with pytest.raises(e.ParamError):
        aerospike.client(config)


def test_setting_rack_aware_and_rack_id():
    config = copy.deepcopy(gconfig)
    config["rack_aware"] = True
    config["rack_id"] = 0x1234
    client = aerospike.client(config)
    assert client is not None


def test_setting_use_services_alternate():
    config = copy.deepcopy(gconfig)
    config["use_services_alternate"] = True
    client = aerospike.client(config)
    assert client is not None


def test_setting_rack_aware_non_bool():
    config = copy.deepcopy(gconfig)
    config["rack_aware"] = "True"
    with pytest.raises(e.ParamError):
        aerospike.client(config)


@pytest.mark.parametrize(
    "rack_id",
    (
        "test_id",  # String
        "3.14",  # Float
        -(1 << 40),  # Too small
        (1 << 32),  # Too large
    ),
)
def test_setting_rack_id_wrong_type(rack_id):
    config = copy.deepcopy(gconfig)
    config["rack_id"] = rack_id
    with pytest.raises(e.ParamError):
        aerospike.client(config)


def test_setting_wrong_type_services_alternate():
    """
    'use_services_alternate' should be a boolean
    """
    config = copy.deepcopy(gconfig)
    config["use_services_alternate"] = "True"
    with pytest.raises(e.ParamError):
        aerospike.client(config)



# Keys
# Only insert two records with the first and second key
keyTuples = [
    ('test', 'demo', '1'),
    ('test', 'demo', '2'),
    ('test', 'demo', '3'),
]
client.put(keyTuples[0], {'bin1': 'value'})
client.put(keyTuples[1], {'bin1': 'value'})

# Check for existence of records using all three keys
keyMetadata = client.exists_many(keyTuples)
print(keyMetadata[0])
print(keyMetadata[1])
print(keyMetadata[2])

# (('test', 'demo', '1', bytearray(...)), {'ttl': 2592000, 'gen': 1})
# (('test', 'demo', '2', bytearray(...)), {'ttl': 2592000, 'gen': 1})
# (('test', 'demo', '3', bytearray(...)), None)



# -*- coding: utf-8 -*-

import pytest
from .as_status_codes import AerospikeStatus
from .test_base_class import TestBaseClass
from aerospike import exception as e
from .index_helpers import ensure_dropped_index


import aerospike


def add_map_keys(client):
    for i in range(5):
        key = ("test", "demo", i)
        rec = {
            "name": "name%s" % (str(i)),
            "addr": "name%s" % (str(i)),
            "numeric_map": {1: 1, 2: 2, 3: 3},
            "string_map": {"sa": "a", "sb": "b", "sc": "c"},
            "age": i,
            "no": i,
        }
        client.put(key, rec)


def remove_map_keys(client):
    for i in range(5):
        key = ("test", "demo", i)
        client.remove(key)


@pytest.mark.usefixtures("connection_with_config_funcs")
class TestMapKeysIndex(object):
    def setup_class(cls):
        cls.connection_setup_functions = [add_map_keys]
        cls.connection_teardown_functions = [remove_map_keys]

    def test_mapkeysindex_with_no_paramters(self):
        """
        Invoke index_mapkeys_create() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_map_keys_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

    def test_mapkeysindex_with_extra_paramters(self):
        """
        Invoke index_mapkeys_create() with extra parameters.
        """
        policy = {}
        with pytest.raises(TypeError):
            self.as_connection.index_map_keys_create(
                "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy, 1
            )

    def test_mapkeysindex_with_correct_parameters(self):
        """
        Invoke index_mapkeys_create() with correct arguments
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_mapkeysindex_with_correct_parameters_numeric(self):
        """
        Invoke index_mapkeys_create() with correct arguments
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_mapkeysindex_with_correct_parameters_numeric_on_stringkeys(self):
        """
        Invoke index_mapkeys_create() with correct arguments
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "string_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_mapkeysindex_with_correct_parameters_string_on_numerickeys(self):
        """
        Invoke index_mapkeys_create() with correct arguments
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "numeric_map", aerospike.INDEX_STRING, "test_numeric_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    @pytest.mark.parametrize(
        "ns, test_set, test_bin, index_name",
        (
            ("a" * 32, "demo", "string_map", "test_string_map"),
            ("test", "a" * 64, "string_map", "test_string_map"),
            ("test", "demo", "a" * 16, "test_string_map"),
            ("test", "demo", "string_map", "a" * 256),
        ),
        ids=("ns too long", "set too long", "bin too long", "index name too long"),
    )
    def test_mapkeys_with_parameters_too_long(self, ns, test_set, test_bin, index_name):
        # Invoke index_map_keys_create() with correct arguments and set
        # length extra
        policy = {}

        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_map_keys_create(ns, test_set, test_bin, aerospike.INDEX_STRING, index_name, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_mapkeysindex_with_incorrect_namespace(self):
        """
        Invoke createindex() with incorrect namespace
        """
        policy = {}
        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_map_keys_create(
                "test1", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
            )

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_mapkeysindex_with_nonexistent_set(self):
        """
        Invoke createindex() with incorrect set
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo1", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_mapkeysindex_on_nonexistent_bin(self):
        """
        Invoke createindex() with nonexistent bin
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "string_map1", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    @pytest.mark.parametrize(
        "ns, test_set, test_bin, index_type, index_name, policy",
        (
            (None, "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", {}),
            (1, "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", {}),
            ("test", 1, "string_map", aerospike.INDEX_STRING, "test_string_map_index", {}),
            ("test", "demo", None, aerospike.INDEX_STRING, "test_string_map_index", {}),
            ("test", "demo", "string_map", None, "test_string_map_index", {}),
            ("test", "demo", "string_map", (), "test_string_map_index", {}),
            ("test", "demo", "string_map", "numeric", "test_string_map_index", {}),
            ("test", "demo", "string_map", aerospike.INDEX_STRING, None, {}),
            ("test", "demo", "string_map", aerospike.INDEX_STRING, 15, {}),
            ("test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", 1),
            ("test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", {"timeout": 0.5}),
        ),
        ids=(
            "Namespace is None",
            "Namespace is Int",
            "Set is int",
            "Bin is None",
            "Index type is None",
            "Index value is tuple",
            "Index type is string",
            "index name is None",
            "Index name is wrong type",
            "Policy is int",
            "Invalid timeout policy",
        ),
    )
    def test_mapkeysindex_invalid_params(self, ns, test_set, test_bin, index_type, index_name, policy):
        """
        Test with various conditions raising paramerrors
        """

        with pytest.raises(e.ParamError):
            self.as_connection.index_map_keys_create(ns, test_set, test_bin, index_type, index_name, policy)
            self.as_connection.index_remove(ns, index_name)
            ensure_dropped_index(self.as_connection, ns, index_name)

    def test_mapkeysindex_with_set_is_none(self):
        """
        Invoke createindex() with set is None, should not raise an error
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", None, "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index")
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_create_same_mapindex_multiple_times(self):
        """
        Invoke createindex() with same arguments
        multiple times on same bin
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )
        assert response_code == AerospikeStatus.AEROSPIKE_OK

        try:
            response_code = self.as_connection.index_map_keys_create(
                "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
            )
        except e.IndexFoundError:
            assert self.server_version < [6, 1]

        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_create_same_mapindex_multiple_times_different_bin(self):
        """
        Invoke createindex with the same name on multiple bins
        """
        policy = {}
        # Create the index the first time
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )
        assert response_code == AerospikeStatus.AEROSPIKE_OK

        # Create a different index with the same name
        with pytest.raises(e.IndexFoundError):
            response_code = self.as_connection.index_map_keys_create(
                "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_string_map_index", policy
            )

        # Cleanup the index
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_create_different_mapindex_multiple_times_same_bin(self):
        """
        Call createindex on the same bin with different names
        """
        policy = {}
        index_names = ("test_string_map_index", "test_string_map_index1")

        # Make sure the indexes don't already exist
        for index_name in index_names:
            try:
                self.as_connection.index_remove("test", index_name, policy)
                ensure_dropped_index(self.as_connection, "test", index_name)
            except Exception:
                pass

        try:
            for index_name in index_names:
                response_code = self.as_connection.index_map_keys_create(
                    "test", "demo", "string_map", aerospike.INDEX_STRING, index_name, policy
                )
                assert response_code == AerospikeStatus.AEROSPIKE_OK
        except e.IndexFoundError:
            assert self.server_version < [6, 1]

        for index_name in index_names:
            try:
                self.as_connection.index_remove("test", index_name, policy)
                ensure_dropped_index(self.as_connection, "test", index_name)
            except Exception:
                pass

    def test_createmapindex_with_policy(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_create_map_integer_index_unicode(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")
        response_code = self.as_connection.index_map_keys_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "uni_age_index", policy
        )

        assert response_code == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "uni_age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")

    def test_mapkeysindex_with_correct_parameters_no_connection(self):
        """
        Invoke index_map_keys_create() with correct arguments no connection
        """
        policy = {}
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.index_map_keys_create(
                "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
            )

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR



# Imports
import aerospike
from aerospike import exception as ex
import sys

# Configure the client
config = {
    'hosts': [ ('127.0.0.1', 3000)]
}

# Create a client and connect it to the cluster
try:
    client = aerospike.client(config).connect()
    client.truncate('test', "demo", 0)
except ex.ClientError as e:
    print("Error: {0} [{1}]".format(e.msg, e.code))
    sys.exit(1)

# Record key tuple: (namespace, set, key)
keyTuple = ('test', 'demo', 'key')


# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass


class TestScanGetPartitionsStatus(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        self.partition_1000_count = 0
        self.partition_1001_count = 0
        self.partition_1002_count = 0
        self.partition_1003_count = 0

        as_connection.truncate(self.test_ns, None, 0)

        for i in range(1, 100000):
            put = 0
            key = (self.test_ns, self.test_set, str(i))
            rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

            if rec_partition == 1000:
                self.partition_1000_count += 1
                put = 1
            if rec_partition == 1001:
                self.partition_1001_count += 1
                put = 1
            if rec_partition == 1002:
                self.partition_1002_count += 1
                put = 1
            if rec_partition == 1003:
                self.partition_1003_count += 1
                put = 1
            if put:
                rec = {
                    "i": i,
                    "s": "xyz",
                    "l": [2, 4, 8, 16, 32, None, 128, 256],
                    "m": {"partition": rec_partition, "b": 4, "c": 8, "d": 16},
                }
                as_connection.put(key, rec)

        def teardown():
            for i in range(1, 100000):
                put = 0
                key = ("test", "demo", str(i))
                rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

                if rec_partition == 1000:
                    self.partition_1000_count += 1
                    put = 1
                if rec_partition == 1001:
                    self.partition_1001_count += 1
                    put = 1
                if rec_partition == 1002:
                    self.partition_1002_count += 1
                    put = 1
                if rec_partition == 1003:
                    self.partition_1003_count += 1
                    put = 1
                if put:
                    as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_scan_get_partitions_status_no_tracking(self):
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        stats = scan_obj.get_partitions_status()
        assert stats == {}

    def test_get_partitions_status_after_foreach(self):
        """
        Resume a scan using foreach.
        """
        records = 0
        resumed_records = 0

        def callback(part_id, input_tuple):
            nonlocal records
            if records == 5:
                return False
            records += 1

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        assert records == 5

        partition_status = scan_obj.get_partitions_status()

        def resume_callback(part_id, input_tuple):
            nonlocal resumed_records
            resumed_records += 1

        scan_obj2 = self.as_connection.scan(self.test_ns, self.test_set)

        policy = {
            "partition_filter": {"begin": 1001, "count": 1, "partition_status": partition_status},
        }

        scan_obj2.foreach(resume_callback, policy)

        assert records + resumed_records == self.partition_1001_count

    def test_scan_get_partitions_status_results(self):
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        # policy = {'partition_filter': {'begin': 1001, 'count': 1}}
        scan_obj.paginate()
        scan_obj.results()

        stats = scan_obj.get_partitions_status()
        assert stats

    def test_scan_get_partitions_status_results_no_tracking(self):
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        # policy = {'partition_filter': {'begin': 1001, 'count': 1}}
        scan_obj.results()

        stats = scan_obj.get_partitions_status()
        assert not stats

    def test_scan_get_partitions_status_results_parts(self):
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        policy = {"partition_filter": {"begin": 1001, "count": 1}}
        results = scan_obj.results(policy)
        assert len(results) == self.partition_1001_count

        stats = scan_obj.get_partitions_status()
        assert stats

    def test_scan_get_partitions_status_foreach_parts(self):
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        ids = []

        def callback(part_id, input_tuple):
            ids.append(part_id)

        policy = {"partition_filter": {"begin": 1001, "count": 1}}
        scan_obj.foreach(callback, policy)
        assert len(ids) == self.partition_1001_count

        stats = scan_obj.get_partitions_status()
        assert stats



import pytest
from aerospike import exception as e

import aerospike


class TestInvalidClientConfig(object):
    def test_no_config(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client()
        assert "No config argument" in err.value.msg

    def test_config_not_dict(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client([])
        assert "Config must be a dict" in err.value.msg

    def test_no_host_in_config(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({})
        assert "Hosts must be a list" in err.value.msg

    def test_wrong_host_type(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": (())})
        assert "Hosts must be a list" in err.value.msg

    def test_empty_host_in_config(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": []})
        assert "Hosts must not be empty" in err.value.msg

    def test_invalid_host_in_list(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": [("localhost", 3000), ()]})
        assert "Invalid host" in err.value.msg

    def test_lua_user_path_too_long(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": [("localhost", 3000)], "lua": {"user_path": "a" * 256}})
        assert "Lua user path too long" in err.value.msg

    def test_non_callable_serializer(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": [("localhost", 3000)], "serialization": (5, lambda x: 5)})
        assert "Serializer must be callable" in err.value.msg

    def test_non_callable_deserializer(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": [("localhost", 3000)], "serialization": (lambda x: 5, 5)})
        assert "Deserializer must be callable" in err.value.msg

    def test_negative_threshold_value(self):
        with pytest.raises(e.ParamError) as err:
            aerospike.client({"hosts": [("localhost", 3000)], "compression_threshold": -1})
        assert "Compression value must not be negative" in err.value.msg

    @pytest.mark.parametrize("policy", ["read", "write", "operate", "batch", "scan", "query", "apply", "remove"])
    @pytest.mark.parametrize(
        "key, value",
        [("total_timeout", "5"), ("socket_timeout", "5"), ("max_retries", "5"), ("sleep_between_retries", "5")],
    )
    def test_invalid_subpolicy_base_types(self, policy, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {policy: subpolicy}})

    @pytest.mark.parametrize("key, value", [("deserialize", "nope"), ("key", "send"), ("replica", "maybe?")])
    def test_invalid_read_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"read": subpolicy}})

    @pytest.mark.parametrize(
        "key, value",
        [
            ("key", "send"),  # should be int
            ("exists", "exists"),  # should be int
            ("gen", "should be a constant integer"),  # should be int
            ("commit_level", "committed"),  # should be int
            ("durable_delete", "durable"),  # should be bool
        ],
    )
    def test_invalid_write_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"write": subpolicy}})

    @pytest.mark.parametrize(
        "key, value",
        [
            ("key", "send"),  # should be int
            ("gen", "should be a constant integer"),  # should be int
            ("replica", "maybe?"),  # should be int
            ("commit_level", "committed"),  # should be int
            ("durable_delete", "durable"),  # should be bool
        ],
    )
    def test_invalid_operat_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"operate": subpolicy}})

    @pytest.mark.parametrize(
        "key, value",
        [
            ("concurrent", "concurrent"),  # should be bool
            ("allow_inline", "False"),  # should be bool
            ("deserialize", "False"),  # should be bool
        ],
    )
    def test_invalid_batch_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"batch": subpolicy}})

    @pytest.mark.parametrize("key, value", [("durable_delete", "durable")])  # should be bool
    def test_invalid_scan_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"scan": subpolicy}})

    #  Keep this parametrized in case query gets additional policies
    @pytest.mark.parametrize(
        "key, value",
        [
            ("deserialize", "False"),  # should be a bool
        ],
    )
    def test_invalid_query_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"query": subpolicy}})

    @pytest.mark.parametrize(
        "key, value",
        [
            # ("gen", "should be a constant integer"),  # gen removed from apply policies by C client 5.0
            ("replica", "maybe?"),  # should be int
            ("commit_level", "committed"),  # should be int
            ("key", "send"),  # should be an int eg. aerospike.POLICY_KEY_SEND
            ("durable_delete", "durable"),  # should be bool
        ],
    )
    def test_invalid_apply_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"apply": subpolicy}})

    @pytest.mark.parametrize(
        "key, value",
        [
            ("key", "send"),  # should be an int eg. aerospike.POLICY_KEY_SEND
            ("durable_delete", "durable"),  # should be bool
            ("gen", "should be a constant integer"),  # should be int
            ("replica", "maybe?"),  # should be int
            ("commit_level", "committed"),  # should be int
        ],
    )
    def test_invalid_remove_policy_types(self, key, value):
        subpolicy = {key: value}
        with pytest.raises(e.ParamError):
            aerospike.client({"hosts": [("localhost", 3000)], "policies": {"remove": subpolicy}})



# -*- coding: utf-8 -*-

import pytest
import time

try:
    from collections import Counter
except ImportError:
    from counter26 import Counter

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


@pytest.mark.usefixtures("as_connection")
class TestExistsMany:
    def test_pos_exists_many_without_policy(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)
        records = self.as_connection.exists_many(self.keys)
        assert isinstance(records, list)
        assert len(records) == rec_length

    def test_pos_exists_many_with_proper_parameters(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)
        records = self.as_connection.exists_many(self.keys, {"max_retries": 1})

        assert isinstance(records, list)
        assert len(records) == rec_length
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4])

    def test_pos_exists_many_with_none_policy(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)

        records = self.as_connection.exists_many(self.keys, None)

        assert isinstance(records, list)
        assert len(records) == rec_length
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4])

    def test_pos_exists_many_with_non_existent_keys(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)

        self.keys.append(("test", "demo", "some_key"))

        records = self.as_connection.exists_many(self.keys)

        assert isinstance(records, list)
        assert len(records) == rec_length + 1
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "some_key"])
        for x in records:
            if x[0][2] == "some_key":
                assert x[1] is None

    def test_pos_exists_many_with_all_non_existent_keys(self):

        keys = [("test", "demo", "NonExistingKey")]

        records = self.as_connection.exists_many(keys)

        assert len(records) == 1
        for x in records:
            if x[0][2] == "key":
                assert x[1] is None

    def test_pos_exists_many_with_initkey_as_digest(self, put_data):

        keys = []
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name1", "age": 1}
        put_data(self.as_connection, key, rec)
        keys.append(key)

        key = ("test", "demo", None, bytearray("ase;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name2", "age": 2}
        put_data(self.as_connection, key, rec)
        keys.append(key)

        records = self.as_connection.exists_many(keys)

        assert isinstance(records, list)
        assert len(records) == 2
        i = 0
        for x in records:
            if i:
                assert x[0][3] == bytearray(b"ase;as[d'as;djk;uyfl")
            else:
                assert x[0][3] == bytearray(b"asd;as[d'as;djk;uyfl")
            i += 1

    def test_pos_exists_many_with_non_existent_keys_in_middle(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)

        self.keys.append(("test", "demo", "some_key"))

        for i in range(15, 20):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, rec)
            self.keys.append(key)

        records = self.as_connection.exists_many(self.keys)

        assert isinstance(records, list)
        assert len(records) == 11
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "some_key", 15, 16, 17, 18, 19])
        for x in records:
            if x[0][2] == "some_key":
                assert x[1] is None

    def test_pos_exists_many_with_record_expiry(self, put_data):
        keys = []
        key = ("test", "demo", 20)
        rec = {"name": "John"}
        meta = {"gen": 3, "ttl": 1}
        put_data(self.as_connection, key, rec, meta)
        keys.append(key)
        time.sleep(2)
        records = self.as_connection.exists_many(keys)
        assert isinstance(records, list)
        assert len(records) == 1
        for record in records:
            assert record[1] is None

    def test_exists_many_with_bytearray_key(self, put_data):
        self.keys = [("test", "demo", bytearray([1, 2, 3]))]
        for key in self.keys:
            put_data(self.as_connection, key, {"byte": "array"})

        records = self.as_connection.exists_many(self.keys)

        bytearray_key = records[0][0]
        assert len(bytearray_key) == 4

        bytearray_pk = bytearray_key[2]
        assert bytearray_pk == bytearray([1, 2, 3])

    # Negative Tests

    def test_neg_exists_many_with_none_keys(self):

        with pytest.raises(e.ParamError):
            self.as_connection.exists_many(None, {})

    def test_neg_exists_many_with_an_invalid_key_in_list(self):

        with pytest.raises(e.ParamError):
            self.as_connection.exists_many([("test", "demo", 1), ("test", "demo", 2), 5])

    def test_neg_exists_many_with_invalid_key(self):

        with pytest.raises(e.ParamError):
            self.as_connection.exists_many("key")

    def test_neg_exists_many_with_invalid_timeout(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)
        policies = {"total_timeout": 0.2}
        with pytest.raises(e.ParamError):
            self.as_connection.exists_many(self.keys, policies)

    def test_neg_exists_many_with_proper_parameters_without_connection(self, put_data):
        self.keys = []
        rec_length = 5
        for i in range(rec_length):
            key = ("test", "demo", i)
            record = {"name": "name%s" % (str(i)), "age": i}
            put_data(self.as_connection, key, record)
            self.keys.append(key)

        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        try:
            client1.exists_many(self.keys, {"total_timeout": 20})

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_exists_many_with_extra_parameter_in_key(self, put_data):
        keys = []
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name1", "age": 1}
        put_data(self.as_connection, key, rec)
        keys.append(key)

        key = ("test", "demo", None, bytearray("ase;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name2", "age": 2}
        put_data(self.as_connection, key, rec)
        keys.append(key)

        keys_get = []
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"), None)
        keys_get.append(key)

        key = ("test", "demo", None, bytearray("ase;as[d'as;djk;uyfl", "utf-8"), None)
        keys_get.append(key)

        try:
            self.as_connection.exists_many(keys_get)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key tuple must be (Namespace, Set, Key) or (Namespace, Set, None, Digest)"

        for key in keys:
            self.as_connection.remove(key)

    def test_neg_exists_many_with_low_timeout(self, put_data):
        keys = []
        key = ("test", "demo", 20)
        rec = {"name": "John"}
        meta = {"gen": 3, "ttl": 1}
        policy = {"timeout": 2}

        try:
            put_data(self.as_connection, key, rec, meta, policy)
        except e.TimeoutError as exception:
            assert exception.code == 9

        keys.append(key)
        records = self.as_connection.exists_many(keys)
        assert isinstance(records, list)
        assert len(records) == 1

    def test_neg_exists_many_with_invalid_ns(self):
        # ToDo: not sure about put operation
        keys = []
        key = ("test2", "demo", 20)
        # self.as_connection.put(key, rec, meta, policy)
        keys.append(key)
        with pytest.raises(e.ClientError):
            self.as_connection.exists_many(keys)

    def test_neg_exists_many_without_any_parameter(self):

        with pytest.raises(TypeError) as typeError:
            self.as_connection.exists_many()

        assert "argument 'keys' (pos 1)" in str(typeError.value)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions import (
    And,
    Eq,
    ListAppend,
    ListAppendItems,
    ListClear,
    ListGetByIndex,
    ListGetByIndexRange,
    ListGetByIndexRangeToEnd,
    ListGetByRank,
    ListGetByRankRange,
    ListGetByRankRangeToEnd,
    ListGetByValue,
    ListGetByValueList,
    ListGetByValueRange,
    ListGetByValueRelRankRange,
    ListGetByValueRelRankRangeToEnd,
    ListInsert,
    ListInsertItems,
    ListRemoveByIndex,
    ListRemoveByIndexRange,
    ListRemoveByIndexRangeToEnd,
    ListRemoveByRank,
    ListRemoveByRankRange,
    ListRemoveByRankRangeToEnd,
    ListRemoveByValue,
    ListRemoveByValueList,
    ListRemoveByValueRange,
    ListRemoveByValueRelRankRange,
    ListRemoveByValueRelRankToEnd,
    ListSet,
    ListSize,
    ListSort,
    Or,
    ResultType,
)

import aerospike

# Constants
_NUM_RECORDS = 9

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}

GEO_POLY = aerospike.GeoJSON(
    {
        "type": "Polygon",
        "coordinates": [
            [
                [-122.500000, 37.000000],
                [-121.000000, 37.000000],
                [-121.000000, 38.080000],
                [-122.500000, 38.080000],
                [-122.500000, 37.000000],
            ]
        ],
    }
)


def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)


def verify_multiple_expression_result(client, test_ns, test_set, expr, op_bin, expected):
    keys = [(test_ns, test_set, i) for i in range(_NUM_RECORDS + 1)]

    # batch get
    res = [rec for rec in client.get_many(keys, policy={"expressions": expr}) if rec[2]]

    assert len(res) == expected


class TestUsrDefinedClass:

    __test__ = False

    def __init__(self, i):
        self.data = i


LIST_BIN_EXAMPLE = [
    None,
    8,
    "string_test" + str(8),
    [26, 27, 28, 8],
    {32: 32, 33: 33, 8: 8, 31: 31},
    bytearray("bytearray_test" + str(8), "utf8"),
    ("bytes_test" + str(8)).encode("utf8"),
    8 % 2 == 1,
    aerospike.null,
    TestUsrDefinedClass(8),
    float(8),
    GEO_POLY,
]


class TestExpressions(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        for i in range(_NUM_RECORDS):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "t": True,
                "age": i,
                "balance": i * 10,
                "key": i,
                "alt_name": "name%s" % (str(i)),
                "list_bin": [
                    None,
                    i,
                    "string_test" + str(i),
                    [26, 27, 28, i],
                    {31: 31, 32: 32, 33: 33, i: i},
                    bytearray("bytearray_test" + str(i), "utf8"),
                    ("bytes_test" + str(i)).encode("utf8"),
                    i % 2 == 1,
                    aerospike.null,
                    TestUsrDefinedClass(i),
                    float(i),
                    GEO_POLY,
                ],
                "ilist_bin": [
                    1,
                    2,
                    6,
                ],
                "slist_bin": ["b", "d", "f"],
                "llist_bin": [[1, 2], [1, 3], [1, 4]],
                "mlist_bin": [{1: 2}, {1: 3}, {1: 4}],
                "bylist_bin": ["b".encode("utf8"), "d".encode("utf8"), "f".encode("utf8")],
                "bolist_bin": [False, False, True],
                "nlist_bin": [None, aerospike.null, aerospike.null],
                "bllist_bin": [TestUsrDefinedClass(1), TestUsrDefinedClass(3), TestUsrDefinedClass(4)],
                "flist_bin": [1.0, 2.0, 6.0],
            }
            self.as_connection.put(key, rec)

        def teardown():
            for i in range(_NUM_RECORDS):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, bin_type, index, return_type, check, expected",
        [
            (None, None, ResultType.INTEGER, 1, aerospike.LIST_RETURN_VALUE, 8, 1),
            (None, None, ResultType.STRING, 2, aerospike.LIST_RETURN_VALUE, "string_test3", 1),
            (None, None, ResultType.BLOB, 6, aerospike.LIST_RETURN_VALUE, "bytes_test3".encode("utf8"), 1),
            (None, None, ResultType.BLOB, 5, aerospike.LIST_RETURN_VALUE, bytearray("bytearray_test3", "utf8"), 1),
            (None, None, ResultType.LIST, 3, aerospike.LIST_RETURN_VALUE, [26, 27, 28, 6], 1),
            ([list_index], [3], ResultType.INTEGER, 3, aerospike.LIST_RETURN_VALUE, 6, 1),
        ],
    )
    def test_list_get_by_index_pos(self, ctx_types, ctx_indexes, bin_type, index, return_type, check, expected):
        """
        Invoke ListGetByIndex().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, p in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, p))
        else:
            ctx = None

        expr = Eq(ListGetByIndex(ctx, return_type, bin_type, index, "list_bin"), check)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, return_type, check, expected",
        [
            (None, None, 8, aerospike.LIST_RETURN_VALUE, [8], 1),
            (None, None, "string_test3", aerospike.LIST_RETURN_VALUE, ["string_test3"], 1),
            (None, None, "bytes_test3".encode("utf8"), aerospike.LIST_RETURN_VALUE, ["bytes_test3".encode("utf8")], 1),
            (
                None,
                None,
                bytearray("bytearray_test3", "utf8"),
                aerospike.LIST_RETURN_VALUE,
                [bytearray("bytearray_test3", "utf8")],
                1,
            ),
            # (None, None, True, aerospike.LIST_RETURN_VALUE, [True], 9)
            # NOTE: this won't work because booleans are not serialized by default in expressions.
            (None, None, None, aerospike.LIST_RETURN_VALUE, [None], _NUM_RECORDS),
            (None, None, [26, 27, 28, 6], aerospike.LIST_RETURN_VALUE, [[26, 27, 28, 6]], 1),
            ([list_index], [3], 6, aerospike.LIST_RETURN_VALUE, [6], 1),
            (
                None,
                None,
                {31: 31, 32: 32, 33: 33, 8: 8},
                aerospike.LIST_RETURN_VALUE,
                [{31: 31, 32: 32, 33: 33, 8: 8}],
                1,
            ),
            (None, None, aerospike.null, aerospike.LIST_RETURN_VALUE, [aerospike.null], _NUM_RECORDS),
            (None, None, GEO_POLY, aerospike.LIST_RETURN_VALUE, [GEO_POLY], _NUM_RECORDS),
            (None, None, TestUsrDefinedClass(4), aerospike.LIST_RETURN_VALUE, [TestUsrDefinedClass(4)], 1),
        ],
    )
    def test_list_get_by_value_pos(self, ctx_types, ctx_indexes, value, return_type, check, expected):
        """
        Invoke ListGetByValue().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Eq(ListGetByValue(ctx, return_type, value, "list_bin"), check)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, begin, end, return_type, check, expected",
        [
            (None, None, 4, 7, aerospike.LIST_RETURN_VALUE, [[4], [5], [6]], 3),
            # (None, None, 5, aerospike.CDTInfinite(), aerospike.LIST_RETURN_COUNT, [10, 10, 10], 4),
            # temporarily failing because of bool jump rank
            # (None, None, 4, 7, aerospike.LIST_RETURN_RANK, [[2], [2], [2]], 3),
            # temporarily failing because of bool jump rank
            # (None, None, 5, aerospike.CDTInfinite(), aerospike.LIST_RETURN_COUNT, [10, 10, 10], 5),
            # temporarily failing because of bool jump rank
            # (None, None, 4, 7, aerospike.LIST_RETURN_RANK, [[1], [1], [1]], 3),
            # temporarily failing because of bool jump rank
            (None, None, "string_test3", "string_test6", aerospike.LIST_RETURN_INDEX, [[2], [2], [2]], 3),
            (
                None,
                None,
                "bytes_test6".encode("utf8"),
                "bytes_test9".encode("utf8"),
                aerospike.LIST_RETURN_COUNT,
                [1, 1, 1],
                3,
            ),
            (
                None,
                None,
                bytearray("bytearray_test3", "utf8"),
                bytearray("bytearray_test6", "utf8"),
                aerospike.LIST_RETURN_REVERSE_INDEX,
                [[6], [6], [6]],
                3,
            ),
            (
                None,
                None,
                [26, 27, 28, 6],
                [26, 27, 28, 9],
                aerospike.LIST_RETURN_VALUE,
                [[[26, 27, 28, 6]], [[26, 27, 28, 7]], [[26, 27, 28, 8]]],
                3,
            ),
            ([list_index], [3], 5, 9, aerospike.LIST_RETURN_REVERSE_RANK, [[3], [3], [3]], 4),
            (
                None,
                None,
                GEO_POLY,
                aerospike.CDTInfinite(),
                aerospike.LIST_RETURN_VALUE,
                [[GEO_POLY], [GEO_POLY], [GEO_POLY]],
                _NUM_RECORDS,
            ),
            (
                None,
                None,
                TestUsrDefinedClass(4),
                TestUsrDefinedClass(7),
                aerospike.LIST_RETURN_VALUE,
                [[TestUsrDefinedClass(4)], [TestUsrDefinedClass(5)], [TestUsrDefinedClass(6)]],
                3,
            ),  # NOTE py_bytes cannot be compard directly server side
        ],
    )
    def test_list_get_by_value_range_pos(self, ctx_types, ctx_indexes, begin, end, return_type, check, expected):
        """
        Invoke ListGetByValueRange().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Or(
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[0]),
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[1]),
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[2]),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "ctx, begin, end, return_type, check, expected",
        [
            ("bad ctx", 10, 13, aerospike.LIST_RETURN_VALUE, [[10], [11], [12]], e.ParamError),
            (None, 10, 13, aerospike.LIST_RETURN_VALUE, [[10], [11], 12], e.InvalidRequest),
        ],
    )
    def test_list_get_by_value_range_neg(self, ctx, begin, end, return_type, check, expected):
        """
        Invoke ListGetByValue() with expected failures.
        """

        expr = Or(
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[0]),
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[1]),
            Eq(ListGetByValueRange(ctx, return_type, begin, end, "list_bin"), check[2]),
        )

        with pytest.raises(expected):
            verify_multiple_expression_result(
                self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
            )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, return_type, check, expected",
        [
            (None, None, [8, [26, 27, 28, 8]], aerospike.LIST_RETURN_VALUE, [8, [26, 27, 28, 8]], 1),
            (None, None, ["string_test3", 3], aerospike.LIST_RETURN_VALUE, [3, "string_test3"], 1),
            (None, None, ["string_test3", 3], aerospike.LIST_RETURN_VALUE, ["string_test3", 3], 0),
            (
                None,
                None,
                ["bytes_test8".encode("utf8"), 8, GEO_POLY],
                aerospike.LIST_RETURN_VALUE,
                [8, "bytes_test8".encode("utf8"), GEO_POLY],
                1,
            ),
            (None, None, LIST_BIN_EXAMPLE, aerospike.LIST_RETURN_VALUE, LIST_BIN_EXAMPLE, 1),
            (None, None, LIST_BIN_EXAMPLE, aerospike.LIST_RETURN_INDEX, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 1),
            (
                None,
                None,
                LIST_BIN_EXAMPLE,
                aerospike.LIST_RETURN_REVERSE_INDEX,
                [11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
                1,
            ),
            (None, None, LIST_BIN_EXAMPLE, aerospike.LIST_RETURN_COUNT, 12, 1),
            # (None, None, [8], aerospike.LIST_RETURN_RANK, [1], 1), temporarily failing because of bool jump rank
            ([list_index], [3], [26, 6], aerospike.LIST_RETURN_INDEX, [0, 3], 1),
        ],
    )
    def test_list_get_by_value_list_pos(self, ctx_types, ctx_indexes, value, return_type, check, expected):
        """
        Invoke ListGetByValueList().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Eq(ListGetByValueList(ctx, return_type, value, "list_bin"), check)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, return_type, check, expected",
        [(None, None, [10, [26, 27, 28, 10]], aerospike.LIST_RETURN_VALUE, (10, [26, 27, 28, 10]), e.InvalidRequest)],
    )
    def test_list_get_by_value_list_neg(self, ctx_types, ctx_indexes, value, return_type, check, expected):
        """
        Invoke ListGetByValueList() with expected failures.
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Eq(ListGetByValueList(ctx, return_type, value, "list_bin"), check)
        with pytest.raises(expected):
            verify_multiple_expression_result(
                self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
            )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, rank, return_type, check, expected",
        [
            ([list_index], [3], 26, 0, aerospike.LIST_RETURN_COUNT, 3, _NUM_RECORDS),
            ([list_index], [3], 7, 1, aerospike.LIST_RETURN_COUNT, 3, 2),
            ([list_index], [3], 7, 2, aerospike.LIST_RETURN_VALUE, [27, 28], 2),
            # (None, None, "string_test8", 0,  aerospike.LIST_RETURN_COUNT, 10, 1),
            # temporarily failing because of bool jump rank
        ],
    )
    def test_list_get_by_value_rel_rank_range_to_end_pos(
        self, ctx_types, ctx_indexes, value, rank, return_type, check, expected
    ):
        """
        Invoke ListGetByValueRelRankRangeToEnd().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Eq(ListGetByValueRelRankRangeToEnd(ctx, return_type, value, rank, "list_bin"), check)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, rank, return_type, expected",
        [([list_index], [3], 26, "bad_rank", "bad_return_type", e.ParamError)],
    )
    def test_list_get_by_value_rel_rank_range_to_end_neg(
        self, ctx_types, ctx_indexes, value, rank, return_type, expected
    ):
        """
        Invoke ListGetByValueRelRankRangeToEnd() with expected failures.
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = ListGetByValueRelRankRangeToEnd(ctx, return_type, value, rank, "list_bin")
        with pytest.raises(expected):
            verify_multiple_expression_result(
                self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
            )

    @pytest.mark.parametrize(
        "ctx_types, ctx_indexes, value, rank, count, return_type, check, expected",
        [
            ([list_index], [3], 26, 0, 3, aerospike.LIST_RETURN_COUNT, 3, _NUM_RECORDS),
            ([list_index], [3], 26, 0, 2, aerospike.LIST_RETURN_VALUE, [27, 26], _NUM_RECORDS),
            (None, None, "string_test10", 0, 1, aerospike.LIST_RETURN_INDEX, [3], 2),
        ],
    )
    def test_list_get_by_value_rel_rank_range_pos(
        self, ctx_types, ctx_indexes, value, rank, count, return_type, check, expected
    ):
        """
        Invoke ListGetByValueRelRankRange().
        """

        if ctx_types is not None:
            ctx = []
            for ctx_type, index in zip(ctx_types, ctx_indexes):
                ctx.append(add_ctx_op(ctx_type, index))
        else:
            ctx = None

        expr = Eq(ListGetByValueRelRankRange(ctx, return_type, value, rank, count, "list_bin"), check)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "list_bin", expected
        )

    @pytest.mark.parametrize(
        "bin, values",
        [
            ("ilist_bin", [ResultType.INTEGER, 6, 1, 7, [2, 6], 1]),
            ("slist_bin", [ResultType.STRING, "f", "b", "g", ["d", "f"], "b"]),
            ("llist_bin", [ResultType.LIST, [1, 4], [1, 2], [1, 6], [[1, 3], [1, 4]], [1, 2]]),
            (
                "bylist_bin",
                [
                    ResultType.BLOB,
                    "f".encode("utf8"),
                    "b".encode("utf8"),
                    "g".encode("utf8"),
                    ["d".encode("utf8"), "f".encode("utf8")],
                    "b".encode("utf8"),
                ],
            ),
            ("flist_bin", [ResultType.FLOAT, 6.0, 1.0, 7.0, [2.0, 6.0], 1.0]),
        ],
    )
    def test_list_read_ops_pos(self, bin, values):
        """
        Invoke various list read expressions with many value types.
        """

        expr = And(
            Eq(
                ListGetByValueRelRankRange(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    ListGetByIndex(None, aerospike.LIST_RETURN_VALUE, values[0], 0, bin),
                    1,
                    3,
                    bin,
                ),
                2,
            ),
            Eq(
                ListGetByValue(
                    None,
                    aerospike.LIST_RETURN_INDEX,
                    values[1],
                    ListGetByValueRange(None, aerospike.LIST_RETURN_VALUE, values[2], values[3], bin),
                ),
                [2],
            ),
            Eq(
                ListGetByValueList(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    values[4],
                    ListGetByValueRelRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, values[5], 1, bin),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    1,
                    ListGetByIndexRange(
                        None,
                        aerospike.LIST_RETURN_VALUE,
                        1,
                        3,
                        bin,
                    ),
                ),
                1,
            ),
            Eq(
                ListGetByRank(
                    None,
                    aerospike.LIST_RETURN_RANK,
                    ResultType.INTEGER,
                    1,  # lets 20 pass with slist_bin
                    ListGetByRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 1, bin),
                ),
                1,
            ),
            Eq(ListGetByRankRange(None, aerospike.LIST_RETURN_COUNT, 1, ListSize(None, bin), bin), 2),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "bin, ctx, policy, values, expected",
        [
            (
                "ilist_bin",
                None,
                {"write_flags": aerospike.LIST_WRITE_ADD_UNIQUE},
                [20, [3, 9], 4, [24, 25], 10, 1, [2, 6], None, 3, 6, 2],
                [[1, 2, 3, 4, 6, 9, 20], [10, 24, 25], [1], []],
            ),
            (
                "slist_bin",
                None,
                {},
                ["h", ["e", "g"], "c", ["x", "y"], "b", "b", ["d", "f"], "b", None, "f", "d"],
                [
                    ["b", "c", "d", "e", "f", "g", "h"],
                    [
                        "b",
                        "x",
                        "y",
                    ],
                    ["b"],
                    ["b", "d"],
                ],
            ),
            (
                "llist_bin",
                None,
                {},
                [
                    [1, 20],
                    [[1, 6], [1, 9]],
                    [1, 5],
                    [[1, 24], [1, 25]],  #
                    [1, 10],
                    [1, 2],  #
                    [[1, 3], [1, 4]],
                    [1, 2],  #
                    [1, 4],
                    [1, 4],
                    [1, 3],  #
                ],
                [[[1, 2], [1, 3], [1, 4], [1, 5], [1, 6], [1, 9], [1, 20]], [[1, 10], [1, 24], [1, 25]], [[1, 2]], []],
            ),
            (
                "mlist_bin",
                None,
                {},
                [
                    {1: 20},
                    [{1: 6}, {1: 9}],
                    {1: 5},
                    [{1: 24}, {1: 25}],  #
                    {1: 10},
                    {1: 2},  #
                    [{1: 3}, {1: 4}],
                    {1: 2},  #
                    {1: 4},
                    {1: 4},
                    {1: 3},  #
                ],
                [[{1: 2}, {1: 3}, {1: 4}, {1: 5}, {1: 6}, {1: 9}, {1: 20}], [{1: 10}, {1: 24}, {1: 25}], [{1: 2}], []],
            ),
            (
                "bylist_bin",
                None,
                {},
                [b"h", [b"e", b"g"], b"c", [b"x", b"y"], b"b", b"b", [b"d", b"f"], b"b", b"e", b"f", b"d"],
                [[b"b", b"c", b"d", b"e", b"f", b"g", b"h"], [b"b", b"x", b"y"], [b"b"], []],
            ),
            (
                "flist_bin",
                None,
                {},
                [20.0, [3.0, 9.0], 4.0, [24.0, 25.0], 10.0, 1.0, [2.0, 6.0], 1.0, 3.0, 6.0, 2.0],
                [[1.0, 2.0, 3.0, 4.0, 6.0, 9.0, 20.0], [10.0, 24.0, 25.0], [1.0], []],
            ),
        ],
    )
    def test_list_mod_ops_pos(self, bin, ctx, policy, values, expected):
        """
        Invoke various list modify expressions with many value types.
        """

        expr = And(
            Eq(
                ListGetByIndexRangeToEnd(
                    ctx,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        ctx,
                        aerospike.LIST_SORT_DEFAULT,
                        ListAppend(
                            ctx,
                            policy,
                            values[0],
                            ListAppendItems(ctx, policy, values[1], ListInsert(ctx, policy, 1, values[2], bin)),
                        ),
                    ),
                ),  # NOTE: invalid on ordered lists
                expected[0],
            ),
            Eq(
                ListSort(
                    ctx,
                    aerospike.LIST_SORT_DEFAULT,
                    ListGetByRankRangeToEnd(
                        ctx,
                        aerospike.LIST_RETURN_VALUE,
                        0,
                        ListInsertItems(
                            ctx, policy, 0, values[3], ListSet(ctx, policy, 0, values[4], ListClear(ctx, bin))
                        ),
                    ),
                ),
                expected[1],
            ),
            Eq(ListRemoveByValue(ctx, values[5], ListRemoveByValueList(ctx, values[6], bin)), []),
            Eq(
                ListRemoveByValueRange(
                    ctx, values[7], values[8], ListRemoveByValueRelRankToEnd(ctx, values[9], 0, bin)
                ),
                expected[3],
            ),
            Eq(ListRemoveByValueRelRankRange(ctx, values[10], 0, 2, ListRemoveByIndex(ctx, 0, bin)), []),
            Eq(ListRemoveByIndexRange(ctx, 0, 1, ListRemoveByIndexRangeToEnd(ctx, 1, bin)), []),
            Eq(ListRemoveByRank(ctx, 0, ListRemoveByRankRangeToEnd(ctx, 1, bin)), []),
            Eq(ListRemoveByRankRange(ctx, 1, 2, bin), expected[2]),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )



# These fail if we don't have a new server
# -*- coding: utf-8 -*-
import pytest
import aerospike
from aerospike import exception as e
from aerospike_helpers.operations import map_operations as map_ops


def skip_less_than_430(version):
    if version < [4, 3]:
        print(version)
        pytest.skip("Requires server > 4.3.0 to work")


class TestMapWriteFlags(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []

        yield

        for key in self.keys:
            self.as_connection.remove(key)

    def test_default_allows_update_and_create(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_DEFAULT}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]
        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "new"
        assert map_bin["new"] == "new"

    def test_create_only_does_not_allow_update(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_CREATE_ONLY}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
        ]
        with pytest.raises(e.AerospikeError):
            self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"

    def test_create_only_allows_create(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_CREATE_ONLY}
        ops = [
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]
        self.as_connection.operate(key, ops)
        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert map_bin["new"] == "new"

    def test_update_only_does_not_allow_create(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_UPDATE_ONLY}
        ops = [
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]

        with pytest.raises(e.AerospikeError):
            self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert "new" not in map_bin

    def test_update_only_allows_update(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_UPDATE_ONLY}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
        ]

        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "new"

    def test_nofail_allows_an_op_to_fail_silently(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": aerospike.MAP_WRITE_FLAGS_UPDATE_ONLY | aerospike.MAP_WRITE_FLAGS_NO_FAIL}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]

        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "new"
        assert "new" not in map_bin

    def test_partial_allows_partial_write(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {
            "map_write_flags": (
                aerospike.MAP_WRITE_FLAGS_CREATE_ONLY
                | aerospike.MAP_WRITE_FLAGS_PARTIAL
                | aerospike.MAP_WRITE_FLAGS_NO_FAIL
            )
        }
        ops = [
            map_ops.map_put_items("map", {"existing": "new", "new1": "new1", "new2": "new2"}, map_policy=map_policy),
        ]

        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert map_bin["new1"] == "new1"
        assert map_bin["new2"] == "new2"

    def test_no_fail_does_not_allow_partial_write(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": (aerospike.MAP_WRITE_FLAGS_CREATE_ONLY | aerospike.MAP_WRITE_FLAGS_NO_FAIL)}
        ops = [
            map_ops.map_put_items("map", {"existing": "new", "new1": "new1", "new2": "new2"}, map_policy=map_policy),
        ]

        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert "new1" not in map_bin
        assert "new2" not in map_bin

    def test_non_int_write_flag_raises_exception(self):
        skip_less_than_430(self.server_version)
        key = "test", "write_flags", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_flags": "waving flag"}
        ops = [
            map_ops.map_put_items("map", {"existing": "new", "new1": "new1", "new2": "new2"}, map_policy=map_policy),
        ]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, ops)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions import (
    And,
    Eq,
    ListBin,
    ListGetByIndexRangeToEnd,
    ListSort,
    MapClear,
    MapGetByIndex,
    MapGetByIndexRange,
    MapGetByIndexRangeToEnd,
    MapGetByKey,
    MapGetByKeyList,
    MapGetByKeyRange,
    MapGetByKeyRelIndexRange,
    MapGetByKeyRelIndexRangeToEnd,
    MapGetByRank,
    MapGetByRankRange,
    MapGetByRankRangeToEnd,
    MapGetByValue,
    MapGetByValueList,
    MapGetByValueRange,
    MapGetByValueRelRankRange,
    MapGetByValueRelRankRangeToEnd,
    MapIncrement,
    MapPut,
    MapPutItems,
    MapRemoveByIndex,
    MapRemoveByIndexRange,
    MapRemoveByIndexRangeToEnd,
    MapRemoveByKey,
    MapRemoveByKeyList,
    MapRemoveByKeyRange,
    MapRemoveByKeyRelIndexRange,
    MapRemoveByKeyRelIndexRangeToEnd,
    MapRemoveByRank,
    MapRemoveByRankRange,
    MapRemoveByRankRangeToEnd,
    MapRemoveByValue,
    MapRemoveByValueList,
    MapRemoveByValueRange,
    MapRemoveByValueRelRankRange,
    MapRemoveByValueRelRankRangeToEnd,
    MapSize,
    ResultType,
)

import aerospike

# Constants
_NUM_RECORDS = 9

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}

GEO_POLY = aerospike.GeoJSON(
    {
        "type": "Polygon",
        "coordinates": [
            [
                [-122.500000, 37.000000],
                [-121.000000, 37.000000],
                [-121.000000, 38.080000],
                [-122.500000, 38.080000],
                [-122.500000, 37.000000],
            ]
        ],
    }
)

GEO_POLY1 = aerospike.GeoJSON(
    {
        "type": "Polygon",
        "coordinates": [
            [
                [-132.500000, 47.000000],
                [-131.000000, 47.000000],
                [-131.000000, 48.080000],
                [-132.500000, 48.080000],
                [-132.500000, 47.000000],
            ]
        ],
    }
)

GEO_POLY2 = aerospike.GeoJSON(
    {
        "type": "Polygon",
        "coordinates": [
            [
                [-132.500000, 47.000000],
                [-131.000000, 47.000000],
                [-131.000000, 48.080000],
                [-132.500000, 48.080000],
                [-132.500000, 80.000000],
            ]
        ],
    }
)


def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)


def verify_multiple_expression_result(client, test_ns, test_set, expr, op_bin, expected):
    keys = [(test_ns, test_set, i) for i in range(_NUM_RECORDS + 1)]

    # batch get
    res = [rec for rec in client.get_many(keys, policy={"expressions": expr}) if rec[2]]

    assert len(res) == expected


class TestUsrDefinedClass:

    __test__ = False

    def __init__(self, i):
        self.data = i


LIST_BIN_EXAMPLE = [
    None,
    8,
    "string_test" + str(8),
    [26, 27, 28, 8],
    {32: 32, 33: 33, 8: 8, 31: 31},
    bytearray("bytearray_test" + str(8), "utf8"),
    ("bytes_test" + str(8)).encode("utf8"),
    8 % 2 == 1,
    aerospike.null,
    TestUsrDefinedClass(8),
    float(8),
    GEO_POLY,
]


class TestExpressions(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        for i in range(_NUM_RECORDS):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "t": True,
                "age": i,
                "balance": i * 10,
                "key": i,
                "alt_name": "name%s" % (str(i)),
                "list_bin": [
                    None,
                    i,
                    "string_test" + str(i),
                    [26, 27, 28, i],
                    {31: 31, 32: 32, 33: 33, i: i},
                    bytearray("bytearray_test" + str(i), "utf8"),
                    ("bytes_test" + str(i)).encode("utf8"),
                    i % 2 == 1,
                    aerospike.null,
                    TestUsrDefinedClass(i),
                    float(i),
                    GEO_POLY,
                ],
                "mlist_bin": [{1: 2}, {1: 3}, {1: 4}],
                "imap_bin": {
                    1: 1,
                    2: 2,
                    3: 6,
                },
                "smap_bin": {"b": "b", "d": "d", "f": "f"},
                "lmap_bin": {1: [1, 2], 2: [1, 3], 3: [1, 4]},
                "mmap_bin": {1: {1: 2}, 2: {1: 3}, 3: {1: 4}},
                "bymap_bin": {1: "b".encode("utf8"), 2: "d".encode("utf8"), 3: "f".encode("utf8")},
                "bomap_bin": {1: False, 2: False, 3: True},
                "nmap_bin": {1: None, 2: aerospike.null, 3: aerospike.null},
                "blmap_bin": {1: TestUsrDefinedClass(1), 2: TestUsrDefinedClass(3), 3: TestUsrDefinedClass(4)},
                "fmap_bin": {1.0: 1.0, 2.0: 2.0, 6.0: 6.0},
                "gmap_bin": {1: GEO_POLY, 2: GEO_POLY1, 3: GEO_POLY2},
            }
            self.as_connection.put(key, rec)

        def teardown():
            for i in range(_NUM_RECORDS):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "bin, values, keys, expected",
        [
            (
                "imap_bin",
                [ResultType.INTEGER, 6, 2, 7, [2, 6], 1, ResultType.INTEGER],
                [3, 2, 4, [2, 3], 2],
                [2, [2, 6], [2, 6]],
            ),
            (
                "smap_bin",
                [ResultType.INTEGER, "f", "d", "g", ["d", "f"], "b", ResultType.STRING],
                ["f", "d", "g", ["d", "f"], "d"],
                ["d", ["d", "f"], ["d", "f"]],
            ),
            (
                "lmap_bin",
                [ResultType.INTEGER, [1, 4], [1, 3], [1, 5], [[1, 3], [1, 4]], [1, 2], ResultType.LIST],
                [3, 2, 4, [2, 3], 2],
                [[1, 3], [[1, 3], [1, 4]], [[1, 3], [1, 4]]],
            ),
        ],
    )
    def test_map_read_ops_pos(self, bin, values, keys, expected):
        """
        Invoke various map read expressions with many value types.
        """

        expr = And(
            Eq(MapGetByKey(None, aerospike.MAP_RETURN_RANK, values[0], keys[0], bin), 2),
            Eq(MapGetByValue(None, aerospike.MAP_RETURN_RANK, values[1], bin), [2]),
            Eq(MapGetByIndex(None, aerospike.MAP_RETURN_RANK, values[0], 1, bin), 1),
            Eq(MapGetByRank(None, aerospike.MAP_RETURN_VALUE, values[6], 1, bin), expected[0]),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByKeyRange(None, aerospike.MAP_RETURN_VALUE, keys[1], keys[2], bin),
                    ),
                ),
                expected[1],
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByKeyList(None, aerospike.MAP_RETURN_INDEX, keys[3], bin),
                    ),
                ),
                [1, 2],
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByKeyRelIndexRangeToEnd(None, aerospike.MAP_RETURN_VALUE, keys[4], 1, bin),
                    ),
                ),
                1,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByKeyRelIndexRange(None, aerospike.MAP_RETURN_VALUE, keys[4], 0, 2, bin),
                    ),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByValueRange(None, aerospike.MAP_RETURN_VALUE, values[2], values[3], bin),
                    ),
                ),
                expected[2],
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByValueList(None, aerospike.MAP_RETURN_INDEX, values[4], bin),
                    ),
                ),
                [1, 2],
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByValueRelRankRangeToEnd(None, aerospike.MAP_RETURN_VALUE, values[5], 1, bin),
                    ),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByValueRelRankRange(None, aerospike.MAP_RETURN_VALUE, values[5], 0, 2, bin),
                    ),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByIndexRangeToEnd(None, aerospike.MAP_RETURN_VALUE, 1, bin),
                    ),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByIndexRange(None, aerospike.MAP_RETURN_VALUE, 1, 2, bin),
                    ),
                ),
                expected[2],
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByRankRangeToEnd(None, aerospike.MAP_RETURN_VALUE, 1, bin),
                    ),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_VALUE,
                    0,
                    ListSort(
                        None,
                        aerospike.LIST_SORT_DEFAULT,
                        MapGetByRankRange(None, aerospike.MAP_RETURN_VALUE, 1, 2, bin),
                    ),
                ),
                expected[2],
            ),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "bin, bin_name, ctx, policy, key, value, expected",
        [
            ("imap_bin", "imap_bin", None, None, 3, 6, [12]),
            ("fmap_bin", "fmap_bin", None, None, 6.0, 6.0, [12.0]),
            (ListBin("mlist_bin"), "mlist_bin", [cdt_ctx.cdt_ctx_list_index(0)], None, 1, 4, [6]),
        ],
    )
    def test_map_increment_pos(self, bin, bin_name, ctx, policy, key, value, expected):
        """
        Invoke MapIncrement() on various integer and float bins.
        """
        expr = Eq(
            MapGetByValue(ctx, aerospike.MAP_RETURN_VALUE, expected[0], MapIncrement(ctx, policy, key, value, bin)),
            expected,
        ).compile()
        verify_multiple_expression_result(self.as_connection, self.test_ns, self.test_set, expr, bin_name, _NUM_RECORDS)

    @pytest.mark.parametrize(
        "bin, ctx, policy, values",
        [
            (
                "imap_bin",
                None,
                {"map_write_flags": aerospike.MAP_WRITE_FLAGS_NO_FAIL},
                [aerospike.CDTInfinite, 10, 1, 1, 3, 6],
            ),
            (
                "smap_bin",
                None,
                {},
                ["j", "j", "b", "b", "f", "f"],
            ),
            (
                "lmap_bin",
                None,
                {},
                [7, [1, 8], 1, [1, 2], 3, [1, 4]],
            ),
            (
                "mmap_bin",
                None,
                {},
                [7, {1: 8}, 1, {1: 2}, 3, {1: 4}],
            ),
            (
                "bymap_bin",
                None,
                {},
                [7, "j".encode("utf8"), 1, "b".encode("utf8"), 3, "f".encode("utf8")],
            ),
            (
                "fmap_bin",
                None,
                {},
                [8.0, 10.0, 1.0, 1.0, 6.0, 6.0],
            ),
        ],
    )
    def test_map_mod_ops_pos(self, bin, ctx, policy, values):
        """
        Invoke various map modify expressions with many value types.
        """

        expr = And(
            Eq(
                MapGetByValue(ctx, aerospike.MAP_RETURN_KEY, values[1], MapPut(ctx, policy, values[0], values[1], bin)),
                [values[0]],
            ),
            Eq(
                MapGetByValue(
                    ctx, aerospike.MAP_RETURN_KEY, values[1], MapPutItems(ctx, policy, {values[0]: values[1]}, bin)
                ),
                [values[0]],
            ),
            # Eq(MapClear(ctx, bin), NOTE: not valid, const map comparison
            #     {1:1}
            # )
            Eq(MapSize(None, MapClear(ctx, bin)), 0),
            Eq(MapGetByValue(ctx, aerospike.MAP_RETURN_COUNT, values[3], MapRemoveByKey(ctx, values[2], bin)), 0),
            Eq(MapGetByValue(ctx, aerospike.MAP_RETURN_COUNT, values[3], MapRemoveByKeyList(ctx, [values[2]], bin)), 0),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByKeyRange(ctx, values[2], values[4], bin),
                ),
                [values[5]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByKeyRelIndexRangeToEnd(ctx, values[2], 1, bin),
                ),
                [values[3]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByKeyRelIndexRange(ctx, values[2], 1, 3, bin),
                ),
                [values[3]],
            ),
            #
            Eq(MapGetByValue(ctx, aerospike.MAP_RETURN_COUNT, values[3], MapRemoveByValue(ctx, values[3], bin)), 0),
            Eq(
                MapGetByValue(ctx, aerospike.MAP_RETURN_COUNT, values[3], MapRemoveByValueList(ctx, [values[3]], bin)),
                0,
            ),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByValueRange(ctx, values[3], values[5], bin),
                ),
                [values[5]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByValueRelRankRangeToEnd(ctx, values[3], 1, bin),
                ),
                [values[3]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx,
                    aerospike.MAP_RETURN_VALUE,
                    values[2],
                    values[0],
                    MapRemoveByValueRelRankRange(ctx, values[3], 1, 3, bin),
                ),
                [values[3]],
            ),
            #
            Eq(MapGetByValue(ctx, aerospike.MAP_RETURN_VALUE, values[3], MapRemoveByIndex(ctx, 0, bin)), []),
            Eq(
                MapGetByKeyRange(
                    ctx, aerospike.MAP_RETURN_VALUE, values[2], values[0], MapRemoveByIndexRange(ctx, 1, 3, bin)
                ),
                [values[3]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx, aerospike.MAP_RETURN_VALUE, values[2], values[0], MapRemoveByIndexRangeToEnd(ctx, 1, bin)
                ),
                [values[3]],
            ),
            #
            Eq(MapGetByValue(ctx, aerospike.MAP_RETURN_VALUE, values[3], MapRemoveByRank(ctx, 0, bin)), []),
            Eq(
                MapGetByKeyRange(
                    ctx, aerospike.MAP_RETURN_VALUE, values[2], values[0], MapRemoveByRankRange(ctx, 1, 3, bin)
                ),
                [values[3]],
            ),
            Eq(
                MapGetByKeyRange(
                    ctx, aerospike.MAP_RETURN_VALUE, values[2], values[0], MapRemoveByRankRangeToEnd(ctx, 1, bin)
                ),
                [values[3]],
            ),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e

from .as_status_codes import AerospikeStatus

try:
    from collections import Counter
except ImportError:
    from counter26 import Counter

from .test_base_class import TestBaseClass

import aerospike

KEY_TYPE_ERROR_MSG = "Keys should be specified as a list or tuple."


def get_primary_key(record):
    """
    Convenience function to extract a primary key from a key tuple
    """
    return record[0][2]


class TestSelectMany(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Add test items before each method
        """
        self.keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"title": "Mr.", "name": "name%s" % (str(i)), "age": i, "addr": "Minisota", "country": "USA"}
            as_connection.put(key, rec)
            self.keys.append(key)
        key = ("test", "demo", "float_value")
        as_connection.put(key, {"float_value": 4.3})
        self.keys.append(key)

        def teardown():
            """
            Remove the items added in setup in order to preserve clean state
            """
            for key in self.keys:
                as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_select_many_without_policy(self):
        """
        Verify that select_many without a policy argument will work
        """
        filter_bins = ["title", "name"]
        records = self.as_connection.select_many(self.keys, filter_bins)

        assert isinstance(records, list)
        assert len(records) == 6
        for k in records:
            bins = k[2].keys()
            #  Verify that only bins specified in filter bins are present
            assert set(bins) <= set(filter_bins)

    def test_select_many_with_unicode_name(self):
        """
        Verify that select_many without a policy argument will work
        """
        filter_bins = ["title", "name"]
        records = self.as_connection.select_many(self.keys, filter_bins)

        assert isinstance(records, list)
        assert len(records) == 6
        for k in records:
            bins = k[2].keys()
            #  Verify that only bins specified in filter bins are present
            assert set(bins) <= set(filter_bins)

    @pytest.mark.skip(reason="This is undocumented behavior")
    def test_select_many_without_policy_binname_tuple(self):
        """
        Verify that select_many without a policy argument will work
        """
        filter_bins = ("title", "name")
        records = self.as_connection.select_many(self.keys, filter_bins)

        assert isinstance(records, list)
        assert len(records) == 6
        for k in records:
            bins = k[2].keys()
            #  Verify that only bins specified in filter bins are present
            assert set(bins) <= set(filter_bins)

    @pytest.mark.parametrize("policy", [{"timeout": 180000}, None], ids=["valid timeout", "None policy"])
    def test_select_many_with_valid_policy_parameters(self, policy):

        filter_bins = ["title", "name", "float_value"]
        records = self.as_connection.select_many(self.keys, filter_bins, policy)

        assert isinstance(records, list)
        assert len(records) == len(self.keys)
        assert Counter([get_primary_key(record) for record in records]) == Counter([key[2] for key in self.keys])
        for record in records:
            bins = record[2].keys()
            assert set(bins) <= set(filter_bins)

    def test_select_many_with_non_existent_keys(self):
        """
        Test that including an invalid key in the list
        of keys does not raise an error
        """
        temp_keys = self.keys[:]
        temp_keys.append(("test", "demo", "non-existent"))

        filter_bins = ["title", "name", "addr"]
        records = self.as_connection.select_many(temp_keys, filter_bins)

        record_counter = Counter([get_primary_key(record) for record in records])
        assert isinstance(records, list)
        assert len(records) == len(temp_keys)

        assert record_counter == Counter([0, 1, 2, 3, 4, "non-existent", "float_value"])
        for record in records:
            if get_primary_key(record) == "non-existent":
                assert record[2] is None
                continue
            bins = record[2].keys()
            assert set(bins) <= set(filter_bins)

    def test_select_many_with_all_non_existent_key(self):
        """
        Test call to select_many with only non existent keys
        """
        keys = [("test", "demo", "key")]
        #  Precalculate the digest of this key
        desired_digest = aerospike.calc_digest(*keys[0])
        filter_bins = ["title", "name", "country"]
        records = self.as_connection.select_many(keys, filter_bins)

        assert len(records) == 1
        assert records == [(("test", "demo", "key", desired_digest), None, None)]

    def test_get_many_with_bytearray_key(self):
        """
        Make sure that get many can handle a a key with a bytearray pk
        """
        keys = [("test", "demo", bytearray([1, 2, 3]))]
        for key in keys:
            self.as_connection.put(key, {"byte": "array"})

        records = self.as_connection.select_many(keys, [])
        self.as_connection.remove(keys[0])

        bytearray_key = records[0][0]
        assert len(bytearray_key) == 4

        bytearray_pk = bytearray_key[2]
        assert bytearray_pk == bytearray([1, 2, 3])

    def test_with_use_batch_direct_true_in_constructor_false_argument(self):

        config = {"policies": {"use_batch_direct": False}}
        client_batch_direct = TestBaseClass.get_new_connection(add_config=config)

        policies = {"use_batch_direct": False}
        records = client_batch_direct.select_many(self.keys, [], policies)
        assert isinstance(records, list)
        assert len(records) == len(self.keys)

        client_batch_direct.close()

    def test_with_use_batch_direct_true_in_constructor(self):

        config = {"policies": {"use_batch_direct": True}}
        client_batch_direct = TestBaseClass.get_new_connection(add_config=config)

        records = self.as_connection.select_many(self.keys, [])
        assert isinstance(records, list)
        assert len(records) == len(self.keys)

        client_batch_direct.close()

    @pytest.mark.skip(reason="This test checks for the same thing twice")
    def test_select_many_with_initkey_as_digest(self):

        keys = []
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name1", "age": 1}
        self.as_connection.put(key, rec)
        keys.append(key)

        key = ("test", "demo", None, bytearray("ase;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name2", "age": 2}
        self.as_connection.put(key, rec)
        keys.append(key)

        records = self.as_connection.select_many(keys, ["name"])

        for key in keys:
            self.as_connection.remove(key)

        assert type(records) == list
        assert len(records) == 2
        i = 0
        for x in records:
            if i:
                assert x[0][3] == bytearray(b"ase;as[d'as;djk;uyfl")
            else:
                assert x[0][3] == bytearray(b"asd;as[d'as;djk;uyfl")
            i += 1

    def test_select_many_with_non_existent_keys_in_middle(self):

        temp_keys = self.keys[:]
        temp_keys.append(("test", "demo", "some_key"))
        default_primary_keys = [k[2] for k in self.keys]
        added_primary_keys = ["some_key"]

        for i in range(15, 20):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i, "position": "Sr. Engineer"}
            self.as_connection.put(key, rec)
            temp_keys.append(key)
            added_primary_keys.append(key[2])

        filter_bins = ["title", "name", "position"]
        records = self.as_connection.select_many(temp_keys, filter_bins)

        for i in range(15, 20):
            key = ("test", "demo", i)
            self.as_connection.remove(key)

        assert isinstance(records, list)
        assert len(records) == (len(default_primary_keys) + len(added_primary_keys))

        # check that the primary keys in the returned records match the
        # expected ones, and occur once each
        return_record_counter = Counter([get_primary_key(record) for record in records])
        assert return_record_counter == Counter(default_primary_keys + added_primary_keys)
        for record in records:
            if get_primary_key(record) == "some_key":
                assert record[2] is None
                continue
            bins = record[2].keys()
            # make sure that only filtered bins are in the returned bins
            # could probably do set(bins) <= set(filter_bins)
            assert set(bins).intersection(set(filter_bins)) == set(bins)

    def test_select_many_with_unicode_bins(self):

        filter_bins = ["title", "name", "country", "addr"]
        records = self.as_connection.select_many(self.keys, filter_bins)

        assert isinstance(records, list)
        assert len(records) == len(self.keys)
        for record in records:
            bins = record[2].keys()
            assert set(bins) <= set(filter_bins)

    def test_select_many_with_empty_bins_list(self):

        records = self.as_connection.select_many(self.keys, [])

        assert isinstance(records, list)
        assert len(records) == len(self.keys)

    # Tests for invalid usage of the function

    @pytest.mark.parametrize(
        "bins",
        (
            ["a", []],
            [[], "a"],
            [None, None],
            [True, "a"],
        ),
    )
    def test_select_many_with_invalid_bin_names_list(self, bins):
        with pytest.raises(e.ParamError):
            self.as_connection.select_many(self.keys, bins)

    def test_select_many_without_any_parameter(self):
        """
        Ensure that calling select_many without parameters
        raises an error
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.select_many()

        assert "argument 'keys' (pos 1)" in str(typeError.value)

    def test_select_many_with_proper_parameters_without_connection(self):

        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        filter_bins = ["title", "name"]

        with pytest.raises(e.ClusterError) as err_info:
            client1.select_many(self.keys, filter_bins)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_select_many_with_invalid_keys(self, invalid_key):
        # invalid_key will be an invalid key_tuple, so we wrap
        # it with a valid key in a list
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.select_many([self.keys[0], invalid_key], ["title"], {})

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_select_many_with_a_single_key(self):
        """
        Verify that passing a single key instead of a list
        raises an error
        """
        key = self.keys[0]
        with pytest.raises(e.ParamError):
            self.as_connection.select_many(key, ["name"], {})

    def test_select_many_with_invalid_timeout(self):

        policies = {"total_timeout": 0.2}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.select_many(self.keys, [], policies)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_select_many_with_an_invalid_key_in_list_batch_direct(self):

        with pytest.raises(e.ParamError):
            self.as_connection.select_many(
                [("test", "demo", 1), ("test", "demo", 2), None], ["title"], {"use_batch_direct": True}
            )

    # Tests for invalid argument types

    @pytest.mark.parametrize("keys_arg", (None, {}, False, "a", 1))
    def test_select_many_with_invalid_keys_type(self, keys_arg):
        with pytest.raises(e.ParamError):
            self.as_connection.select_many(keys_arg, ["name"], {})

    @pytest.mark.parametrize("bins", (None, {}, False, "a", 1))
    def test_select_many_with_invalid_bins_arg_type(self, bins):
        with pytest.raises(e.ParamError):
            self.as_connection.select_many(self.keys, bins)

    @pytest.mark.parametrize("policy", ([], False, 1, "policy"))
    def test_select_many_with_invalid_policy_arg_type(self, policy):
        bins = ["name"]
        with pytest.raises(e.ParamError):
            self.as_connection.select_many(self.keys, bins, policy)



# Insert a record
client.put(keyTuple, {"bin1": 4})

# Try to remove it with the wrong generation
try:
    client.remove(keyTuple, meta={'gen': 5}, policy={'gen': aerospike.POLICY_GEN_EQ})
except ex.AerospikeError as e:
    # Error: AEROSPIKE_ERR_RECORD_GENERATION [3]
    print("Error: {0} [{1}]".format(e.msg, e.code))

# Remove it ignoring generation
client.remove(keyTuple)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike_helpers.expressions import (
    BitAdd,
    BitAnd,
    BitCount,
    BitGet,
    BitGetInt,
    BitInsert,
    BitLeftScan,
    BitLeftShift,
    BitNot,
    BitOr,
    BitRemove,
    BitResize,
    BitRightScan,
    BitRightShift,
    BitSet,
    BitSetInt,
    BitSubtract,
    BitXor,
    Eq,
)

import aerospike

# Constants
_NUM_RECORDS = 9

GEO_POLY = aerospike.GeoJSON(
    {
        "type": "Polygon",
        "coordinates": [
            [
                [-122.500000, 37.000000],
                [-121.000000, 37.000000],
                [-121.000000, 38.080000],
                [-122.500000, 38.080000],
                [-122.500000, 37.000000],
            ]
        ],
    }
)


def verify_multiple_expression_result(client, test_ns, test_set, expr, op_bin, expected):
    keys = [(test_ns, test_set, i) for i in range(_NUM_RECORDS + 1)]

    # batch get
    res = [rec for rec in client.get_many(keys, policy={"expressions": expr}) if rec[2]]

    assert len(res) == expected


class TestUsrDefinedClass:

    __test__ = False

    def __init__(self, i):
        self.data = i


class TestExpressions(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        for i in range(_NUM_RECORDS):
            key = ("test", "demo", i)
            rec = {"1bits_bin": bytearray([1] * 8)}
            self.as_connection.put(key, rec)

        def teardown():
            for i in range(_NUM_RECORDS):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "policy, bytes_size, flags, bin, expected",
        [
            (None, 10, None, "1bits_bin", bytearray([1])),
            ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY}, 10, None, "1bits_bin", bytearray([1])),
            (None, 10, aerospike.BIT_RESIZE_FROM_FRONT, "1bits_bin", bytearray([0])),
        ],
    )
    def test_bit_resize_pos(self, policy, bytes_size, flags, bin, expected):
        """
        Test BitResize expression.
        """

        expr = Eq(BitGet(8, 8, BitResize(policy, bytes_size, flags, bin)), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "policy, byte_offset, byte_size, bin, expected",
        [
            (None, 0, 1, "1bits_bin", bytearray([0] * 1)),
            ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY}, 0, 1, "1bits_bin", bytearray([0] * 1)),
        ],
    )
    def test_bit_remove_ops_pos(self, policy, byte_offset, byte_size, bin, expected):
        """
        Test BitRemove expression.
        """

        expr = Eq(BitRemove(policy, byte_offset, byte_size, bin), bytearray([1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_BitInsert_pos(self, policy):
        """
        Test BitInsert expression.
        """

        expr = Eq(BitInsert(policy, 1, bytearray([3]), "1bits_bin"), bytearray([1, 3, 1, 1, 1, 1, 1, 1, 1]))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_set_pos(self, policy):
        """
        Test BitSet expression.
        """

        expr = Eq(
            BitSet(policy, 7, 1, bytearray([255]), BitSet(policy, 0, 8 * 8, bytearray([0] * 8), "1bits_bin")),
            bytearray([1] + [0] * 7),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_BitOr_pos(self, policy):
        """
        Test BitOr expression.
        """

        expr = Eq(BitOr(policy, 0, 8, bytearray([8]), "1bits_bin"), bytearray([9] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_BitXor_pos(self, policy):
        """
        Test BitXor expression.
        """

        expr = Eq(BitXor(policy, 0, 8, bytearray([1]), "1bits_bin"), bytearray([0] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_and_pos(self, policy):
        """
        Test BitAnd expression.
        """

        expr = Eq(BitAnd(policy, 0, 8, bytearray([0]), "1bits_bin"), bytearray([0] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_not_pos(self, policy):
        """
        Test BitNot expression.
        """

        expr = Eq(BitNot(policy, 0, 64, "1bits_bin"), bytearray([254] * 8))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_left_shift_pos(self, policy):
        """
        Test BitLeftShift expression.
        """

        expr = Eq(BitLeftShift(policy, 0, 8, 3, "1bits_bin"), bytearray([8] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_right_shift_pos(self, policy):
        """
        Test BitRightShift expression.
        """

        expr = Eq(BitRightShift(policy, 0, 8, 1, BitLeftShift(None, 0, 8, 3, "1bits_bin")), bytearray([4] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "policy, bit_offset, bit_size, value, action, bin, expected",
        [
            (None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL, "1bits_bin", bytearray([1] + [2] + [1] * 6)),
            (
                {"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY},
                8,
                8,
                1,
                aerospike.BIT_OVERFLOW_FAIL,
                "1bits_bin",
                bytearray([1] + [2] + [1] * 6),
            ),
        ],
    )
    def test_bit_add_pos(self, policy, bit_offset, bit_size, value, action, bin, expected):
        """
        Test BitAdd expression.
        """

        expr = Eq(BitAdd(policy, bit_offset, bit_size, value, action, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "policy, bit_offset, bit_size, value, action, bin, expected",
        [
            (None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL, "1bits_bin", bytearray([1] + [0] + [1] * 6)),
            (
                {"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY},
                8,
                8,
                1,
                aerospike.BIT_OVERFLOW_FAIL,
                "1bits_bin",
                bytearray([1] + [0] + [1] * 6),
            ),
        ],
    )
    def test_bit_subtract_pos(self, policy, bit_offset, bit_size, value, action, bin, expected):
        """
        Test BitSubtract expression.
        """

        expr = Eq(BitSubtract(policy, bit_offset, bit_size, value, action, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("policy", [(None), ({"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY})])
    def test_bit_set_int_pos(self, policy):
        """
        Test BitSetInt expression.
        """

        expr = Eq(BitSetInt(policy, 7, 1, 0, "1bits_bin"), bytearray([0] + [1] * 7))

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), "1bits_bin", _NUM_RECORDS
        )

    @pytest.mark.parametrize("bit_offset, bit_size, bin, expected", [(8, 8, "1bits_bin", bytearray([1]))])
    def test_bit_get_pos(self, bit_offset, bit_size, bin, expected):
        """
        Test BitGet expression.
        """

        expr = Eq(BitGet(bit_offset, bit_size, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bit_offset, bit_size, bin, expected", [(16, 8 * 3, "1bits_bin", 3)])
    def test_bit_count_pos(self, bit_offset, bit_size, bin, expected):
        """
        Test BitCount expression.
        """

        expr = Eq(BitCount(bit_offset, bit_size, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bit_offset, bit_size, value, bin, expected", [(0, 8, True, "1bits_bin", 7)])
    def test_bit_left_scan_pos(self, bit_offset, bit_size, value, bin, expected):
        """
        Test BitLeftScan expression.
        """

        expr = Eq(BitLeftScan(bit_offset, bit_size, value, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bit_offset, bit_size, value, bin, expected", [(0, 8, True, "1bits_bin", 7)])
    def test_bit_right_scan_pos(self, bit_offset, bit_size, value, bin, expected):
        """
        Test BitRightScan expression.
        """

        expr = Eq(BitRightScan(bit_offset, bit_size, value, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bit_offset, bit_size, bin, expected", [(0, 8, "1bits_bin", 1)])
    def test_bit_get_int_pos(self, bit_offset, bit_size, bin, expected):
        """
        Test BitGetInt expression.
        """

        expr = Eq(BitGetInt(bit_offset, bit_size, True, bin), expected)

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function

import aerospike
import sys

from optparse import OptionParser

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-i", "--partition_id", dest="partition", type="int", default=0,
    help="Partition id from where to scan.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)]
}

##########################################################################
# Application
##########################################################################

exitCode = 0

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None

        s = client.scan(namespace, set)

        partition_policy = None

        if options.partition > 0:
            # project specified bins
            partition_policy = {'partition_filter': {'begin': options.partition, 'count': 1}}
            print(f'partition_id: {options.partition}')
    
        records = []

        # callback to be called for each record read
        def callback(input_tuple):
            (_, _, record) = input_tuple
            records.append(record)
            print(record)
        
        client.truncate('test', "demo", 0)

        # invoke the operations, and for each record invoke the callback
        s.foreach(callback, partition_policy)
        existing_count = len(records)
        if existing_count > 0:
            print(f"{existing_count} records are exist already in partition:{options.partition}.")

        count = 0
        for i in range(1, 80000):
            rec_partition = client.get_key_partition_id('test', 'demo', str(i))

            if rec_partition == options.partition: # and not client.exists(('test', 'demo', str(i))):
                
                count = count + 1
                rec = {
                    'i': i,
                    's': 'xyz',
                    'l': [2, 4, 8, 16, 32, None, 128, 256],
                    'm': {'partition': rec_partition, 'b': 4, 'c': 8, 'd': 16}
                }
                client.put(('test', 'demo', str(i)), rec)
        
        records.clear()
        # invoke the operations, and for each record invoke the callback
        s.foreach(callback, partition_policy)

        print("---")
        print(f"{count} records are put into partition:{options.partition}.")
        print(f"{len(records)} records are found in partition:{options.partition}.")
 
    except Exception as e:
        print("error: {0}".format(e), file=sys.stderr)
        rc = 1

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3

##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e
from .index_helpers import ensure_dropped_index
from aerospike_helpers import cdt_ctx

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}


def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)


ctx_list_index = []
ctx_list_index.append(add_ctx_op(list_index, 0))

ctx_list_rank = []
ctx_list_rank.append(add_ctx_op(list_rank, -1))

ctx_list_value = []
ctx_list_value.append(add_ctx_op(list_value, 3))

ctx_map_index = []
ctx_map_index.append(add_ctx_op(map_index, 0))

ctx_map_key = []
ctx_map_key.append(add_ctx_op(map_key, "sb"))

ctx_map_rank = []
ctx_map_rank.append(add_ctx_op(map_rank, -1))

ctx_map_value = []
ctx_map_value.append(add_ctx_op(map_value, 3))


class TestCDTIndex(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            pytest.skip("It only applies to >= 6.1 enterprise edition")

        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "addr": "name%s" % (str(i)),
                "numeric_list": [1, 2, 3, 4],
                "string_list": ["a", "b", "c", "d"],
                "geojson_list": [
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.096449, 37.421868]}),
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.053321, 37.434212]}),
                ],
                "numeric_map": {"a": 1, "b": 2, "c": 3},
                "string_map": {"sa": "a", "sb": "b", "sc": "c"},
                "age": i,
                "no": i,
            }
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def test_pos_cdtindex_with_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_info_command(self):
        """
        Invoke index_cdt_create() with info command
        """
        policy = {}

        bs_b4_cdt = self.as_connection.get_cdtctx_base64(ctx_list_index)

        r = []
        r.append("sindex-create:ns=test;set=demo;indexname=test_string_list_cdt_index")
        r.append(";indextype=%s" % (cdt_ctx.index_type_string(aerospike.INDEX_TYPE_LIST)))
        r.append(";indexdata=string_list,%s" % (cdt_ctx.index_datatype_string(aerospike.INDEX_STRING)))
        r.append(";context=%s" % (bs_b4_cdt))
        req = "".join(r)

        # print("req is ==========={}", req)
        retobj = self.as_connection.info_all(req, policy=None)
        # print("res is ==========={}", res)

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj != 0

    def test_pos_cdtindex_with_listrank_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_rank},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_listvalue_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_value},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_mapindex_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_MAPKEYS,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_map_index},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_mapvalue_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_MAPVALUES,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_map_value},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_maprankvalue_correct_parameters(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_MAPVALUES,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_map_rank},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    # TODO: duplicate test name
    def test_pos_cdtindex_with_correct_parameters1(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_MAPVALUES,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_map_rank},
            policy,
        )

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj == 0

    def test_pos_cdtindex_with_correct_parameters_numeric(self):
        """
        Invoke index_cdt_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "numeric_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_NUMERIC,
            "test_numeric_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_cdt_index")

    def test_pos_cdtindex_with_correct_parameters_set_length_extra(self):
        # Invoke index_cdt_create() with correct arguments and set length
        # extra
        set_name = "a"
        for _ in range(100):
            set_name = set_name + "a"
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                "test",
                set_name,
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                "test_string_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )
            assert False
        except e.InvalidRequest as exception:
            assert exception.code == 4
        except Exception as exception:
            assert isinstance(exception, e.InvalidRequest)

    def test_pos_cdtindex_with_incorrect_bin(self):
        """
        Invoke createindex() with incorrect bin
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list1",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

    def test_pos_create_same_cdtindex_multiple_times(self):
        """
        Invoke createindex() with multiple times on same bin
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "numeric_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_NUMERIC,
            "test_numeric_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )
        if retobj == 0:
            try:
                self.as_connection.index_cdt_create(
                    "test",
                    "demo",
                    "numeric_list",
                    aerospike.INDEX_TYPE_LIST,
                    aerospike.INDEX_NUMERIC,
                    "test_numeric_list_cdt_index",
                    {"ctx": ctx_list_index},
                    policy,
                )
            except e.IndexFoundError:
                assert self.server_version < [6, 1]
            self.as_connection.index_remove("test", "test_numeric_list_cdt_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_numeric_list_cdt_index")
        else:
            assert False

    def test_pos_create_same_cdtindex_multiple_times_different_bin(self):
        """
        Invoke createindex() with multiple times on different bin
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )
        if retobj == 0:
            with pytest.raises(e.IndexFoundError):
                retobj = self.as_connection.index_cdt_create(
                    "test",
                    "demo",
                    "numeric_list",
                    aerospike.INDEX_TYPE_LIST,
                    aerospike.INDEX_NUMERIC,
                    "test_string_list_cdt_index",
                    {"ctx": ctx_list_index},
                    policy,
                )
                self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
                ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

            self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")
        else:
            assert True is False

    def test_pos_create_different_cdtindex_multiple_times_same_bin(self):
        """
                    Invoke createindex() with multiple times on same bin with different
        name
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )
        if retobj == 0:
            try:
                retobj = self.as_connection.index_cdt_create(
                    "test",
                    "demo",
                    "string_list",
                    aerospike.INDEX_TYPE_LIST,
                    aerospike.INDEX_STRING,
                    "test_string_list_cdt_index1",
                    {"ctx": ctx_list_index},
                    policy,
                )
            except e.IndexFoundError:
                assert self.server_version < [6, 1]

            self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")
        else:
            assert True is False

    def test_pos_createcdtindex_with_policy(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 10000}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "num_list_pol",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_NUMERIC,
            "test_numeric_list_cdt_index_pol",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_cdt_index_pol", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_cdt_index_pol")

    def test_pos_createcdtindex_with_policystring(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 10000}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "test_string_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

    """
    This test case causes a db crash and hence has been commented. Work pending
on the C-client side
    def test_createindex_with_long_index_name(self):
            Invoke createindex() with long index name
        policy = {}
        retobj = self.as_connection.index_cdt_create( 'test', 'demo',
'age',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453rf9qwfa\
sdcfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc', {'ctx': ctx_list_index}, policy)

        assert retobj == 0L
        self.as_connection.index_remove(policy, 'test',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453rf9qwfasd\
cfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc');
    """

    def test_pos_create_liststringindex_unicode_positive(self):
        """
        Invoke create string index() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "string_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_STRING,
            "uni_name_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "uni_name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_name_index")

    def test_pos_create_list_integer_index_unicode(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo",
            "numeric_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_NUMERIC,
            "uni_age_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "uni_age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")

    # def test_pos_create_list_geojson_index(self):
    #     """
    #         Invoke createindex() with correct arguments
    #     """
    #     policy = {}
    #     retobj = self.as_connection.index_cdt_create(
    #         'test', 'demo', 'geojson_list', aerospike.INDEX_TYPE_LIST,
    #         aerospike.INDEX_GEO2DSPHERE,
    #         'geo_index', {'ctx': ctx_list_index}, policy)

    #     assert retobj == 0
    #     self.as_connection.index_remove('test', 'geo_index', policy)
    #     ensure_dropped_index(self.as_connection, 'test', 'geo_index')

    # Negative tests
    def test_neg_cdtindex_with_namespace_is_none(self):
        """
        Invoke createindex() with namespace is None
        """
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                None,
                "demo",
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                "test_string_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Namespace should be a string"

    def test_neg_cdtindex_with_set_is_int(self):
        """
        Invoke createindex() with set is int
        """
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                "test",
                1,
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                "test_string_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )
            assert False
        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Set should be string, unicode or None"
        except Exception as exception:
            assert isinstance(exception, e.ParamError)

    def test_neg_cdtindex_with_set_is_none(self):
        """
        Invoke createindex() with set is None
        """
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                "test",
                None,
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                "test_string_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Set should be a string"
        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

    def test_neg_cdtindex_with_bin_is_none(self):
        """
        Invoke createindex() with bin is None
        """
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                "test",
                "demo",
                None,
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_NUMERIC,
                "test_numeric_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin should be a string"

    def test_neg_cdtindex_with_index_is_none(self):
        """
        Invoke createindex() with index is None
        """
        policy = {}
        try:
            self.as_connection.index_cdt_create(
                "test",
                "demo",
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                None,
                {"ctx": ctx_list_index},
                policy,
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Index name should be string or unicode"

    def test_neg_cdtindex_with_incorrect_namespace(self):
        """
        Invoke createindex() with incorrect namespace
        """
        policy = {}

        try:
            self.as_connection.index_cdt_create(
                "test1",
                "demo",
                "numeric_list",
                aerospike.INDEX_TYPE_DEFAULT,
                aerospike.INDEX_NUMERIC,
                "test_numeric_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )

        except e.InvalidRequest as exception:
            assert exception.code == 4

    def test_neg_cdtindex_with_incorrect_set(self):
        """
        Invoke createindex() with incorrect set
        """
        policy = {}
        retobj = self.as_connection.index_cdt_create(
            "test",
            "demo1",
            "numeric_list",
            aerospike.INDEX_TYPE_LIST,
            aerospike.INDEX_NUMERIC,
            "test_numeric_list_cdt_index",
            {"ctx": ctx_list_index},
            policy,
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_cdt_index")

    def test_neg_cdtindex_with_correct_parameters_no_connection(self):
        """
        Invoke index_cdt_create() with correct arguments no connection
        """
        policy = {}
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        try:
            client1.index_cdt_create(
                "test",
                "demo",
                "string_list",
                aerospike.INDEX_TYPE_LIST,
                aerospike.INDEX_STRING,
                "test_string_list_cdt_index",
                {"ctx": ctx_list_index},
                policy,
            )

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_cdtindex_with_no_paramters(self):
        """
        Invoke index_cdt_create() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_cdt_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("as_connection")
class TestGetNodeNames(object):
    """
    Test Cases for the use of aerospike.Client.get_node_names
    method
    """

    def test_pos_get_node_names(self):
        """
        Test successful to aerospike.Client.get_node_names
        without parameters.

        It should return something like [{'address': '1.1.1.1', 'port': 3000, 'node_name': 'BCER199932C'}]
        """
        response = self.as_connection.get_node_names()
        assert response is not None
        assert isinstance(response, list)
        assert isinstance(response[0], dict)
        assert isinstance(response[0]["address"], str)
        assert isinstance(response[0]["port"], int)
        assert isinstance(response[0]["node_name"], str)
        assert len(response[0]) == 3

    # Tests for behaviors that raise errors
    def test_pos_get_node_names_without_connection(self):
        """
        Test that an attempt to call get_node_names before a connection
        is established will raise the expected error
        """
        config = TestBaseClass.get_connection_config()
        unconnected_client = aerospike.client(config)
        unconnected_client.close()

        try:
            unconnected_client.get_node_names()

        except e.ClusterError as exception:
            assert exception.code == 11
            assert exception.msg == "No connection to aerospike cluster."



# -*- coding: utf-8 -*-

from .test_base_class import TestBaseClass

import aerospike


class TestIsConnected(object):
    def setup_class(cls):
        """
        Setup the config which is used in the tests
        """
        cls.config = TestBaseClass.get_connection_config()

    def _connect(self):
        """
        Private method to connect self.client to the server
        with or without username and password
        """
        # if self.user is None and self.password is None:
        #     self.client = aerospike.client(self.config).connect()
        # else:
        #     self.client = aerospike.client(self.config).connect(self.user,
        #                                                         self.password)
        self.client = TestBaseClass.get_new_connection()

    def test_is_connected_before_connect(self):
        """
        Client call itself establishes connection.
        """
        client = aerospike.client(self.config)
        assert client.is_connected() is True

    def test_pos_is_connected(self):
        """
        Test that is_connected returns True after a connection is established
        """
        self._connect()  # Establish a connection
        assert self.client.is_connected() is True
        self.client.close()  # Close the connection

    def test_pos_is_connected_after_multiple_connects(self):
        """
        Test that is_connected returns True after a connection is established
        """
        self._connect()  # Establish a connection
        self.client.connect()
        self.client.connect()
        assert self.client.is_connected() is True
        self.client.close()  # Close the connection

    def test_is_connected_after_close(self):
        """
        Client call itself establishes connection.
        Connect/Close are deprecated and it is no-op to client
        """
        self._connect()
        assert self.client.is_connected() is True
        self.client.close()
        assert self.client.is_connected() is False
